package java.util.concurrent.atomic;

import sun.misc.Unsafe;

public class AtomicStampedReference<V>
{
    private volatile AtomicStampedReference.Pair<V> pair;
    private static final Unsafe UNSAFE;
    private static final long pairOffset;
    
    public AtomicStampedReference(V v, int n) {
        this.pair = AtomicStampedReference.Pair.of(v, n);
    }
    
    public V getReference() {
        return this.pair.reference;
    }
    
    public int getStamp() {
        return this.pair.stamp;
    }
    
    public V get(int[] array) {
        AtomicStampedReference.Pair<Object> pair = this.pair;
        array[0] = pair.stamp;
        return pair.reference;
    }
    
    public boolean weakCompareAndSet(V v, V v2, int n, int n2) {
        return this.compareAndSet(v, v2, n, n2);
    }
    
    public boolean compareAndSet(V v, V v2, int n, int n2) {
        AtomicStampedReference.Pair<V> pair = this.pair;
        return v == pair.reference && n == pair.stamp && ((v2 == pair.reference && n2 == pair.stamp) || this.casPair(pair, AtomicStampedReference.Pair.of(v2, n2)));
    }
    
    public void set(V v, int n) {
        AtomicStampedReference.Pair<Object> pair = this.pair;
        if (v != pair.reference || n != pair.stamp) {
            this.pair = AtomicStampedReference.Pair.of(v, n);
        }
    }
    
    public boolean attemptStamp(V v, int n) {
        AtomicStampedReference.Pair<V> pair = this.pair;
        return v == pair.reference && (n == pair.stamp || this.casPair(pair, AtomicStampedReference.Pair.of(v, n)));
    }
    
    private boolean casPair(AtomicStampedReference.Pair<V> expected, AtomicStampedReference.Pair<V> x) {
        return AtomicStampedReference.UNSAFE.compareAndSwapObject(this, AtomicStampedReference.pairOffset, expected, x);
    }
    
    static long objectFieldOffset(Unsafe unsafe, String s, Class<?> clazz) {
        try {
            return unsafe.objectFieldOffset(clazz.getDeclaredField(s));
        }
        catch (NoSuchFieldException cause) {
            NoSuchFieldError noSuchFieldError = new NoSuchFieldError(s);
            noSuchFieldError.initCause(cause);
            throw noSuchFieldError;
        }
    }
    
    static {
        UNSAFE = Unsafe.getUnsafe();
        pairOffset = objectFieldOffset(AtomicStampedReference.UNSAFE, "pair", AtomicStampedReference.class);
    }
    
    private static class Pair<T>
    {
        final T reference;
        final int stamp;
        
        private Pair(T reference, int stamp) {
            this.reference = reference;
            this.stamp = stamp;
        }
        
        static <T> AtomicStampedReference.Pair<T> of(T t, int n) {
            return new AtomicStampedReference.Pair(t, n);
        }
    }
}
